import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
import h5py
import sklearn.metrics
import math
import scipy.stats

def get_r_squared(pred, truth):
    curr_r_squared = []
    for i in range(pred.shape[0]):
        x = pred[i,:]
        y = truth[i,:]
        _, _, r_squared, _, _ = scipy.stats.mstats.linregress(x, y)
        curr_r_squared.append(r_squared)
    curr_r_squared_avg = np.mean(curr_r_squared)
    return curr_r_squared_avg

def compute_loss(model, x, gt_behaviour, train_behaviour_decoder, latent_dim, batch_size, rnn_units,num_factors, reverse_gradient=False, training=False):
  z = np.zeros((x.shape[0], latent_dim))
  for chunk in range(x.shape[0]//batch_size):             
     mean, logvar = model.encode(x[chunk*batch_size: (chunk + 1)*(batch_size)], reverse_gradient = reverse_gradient, training=training)
     z[chunk*batch_size: (chunk + 1)*(batch_size)] = model.reparameterize(mean, logvar)

  z_all = [z]
  for t in range(model.num_steps-1):
      z_all.append(tf.zeros((z.shape)))
  z_all = tf.stack(z_all, axis = 1)
  
  x_logit = np.zeros((x.shape[0], x.shape[1], rnn_units*2))
  for chunk in range(x.shape[0]//batch_size):             
     x_logit[chunk*batch_size: (chunk + 1)*(batch_size)] = model.decode(z_all[chunk*batch_size: (chunk + 1)*(batch_size)], training =training)

  total_recon_loss = 0
  for t in range(x.shape[1]):
       loglikelihood_t = Poisson(model.rates(x_logit[:,t,:], training=training)).logp(x[:,t,:])
       total_recon_loss += tf.reduce_sum(loglikelihood_t, axis = [1])
  logpx_z = -tf.reduce_mean(total_recon_loss)
  
  
  behaviour_preds = np.zeros((x.shape[0], x.shape[1], num_factors))
  for chunk in range(x.shape[0]//batch_size):                 
     behaviour_preds[chunk*batch_size: (chunk + 1)*(batch_size)] = model.behaviour_decoder(z_all[chunk*batch_size: (chunk + 1)*(batch_size)], training =training)

  total_b_diff = 0
  for t in range(x.shape[1]):
         diff_t = tf.square(tf.cast(gt_behaviour[:,t,:], tf.float32) - model.behaviour_factor_decoder(behaviour_preds[:,t,:], training=training))
         total_b_diff += tf.reduce_sum(diff_t, axis = [1])
         
  fin_predicted_behaviour = np.zeros_like(gt_behaviour)
  for t in range(x.shape[1]):
        fin_predicted_behaviour[:,t,:] =  model.behaviour_factor_decoder(behaviour_preds[:,t,:], training=training)
  r2 = get_r_squared(fin_predicted_behaviour, gt_behaviour)             
  behaviour_loss = tf.reduce_mean(total_b_diff)

  kl_loss = - 0.5 * tf.reduce_mean(logvar - tf.square(mean) - tf.exp(logvar) + 1)                  

  if train_behaviour_decoder:
      full_loss = behaviour_loss+ 1.0*kl_loss +logpx_z
  else:      
      full_loss = 0*behaviour_loss#  +logpx_z + kl_loss
  return full_loss, logpx_z, behaviour_loss, kl_loss, r2

def train_behaviour(model, x, gt_behaviour, reverse_gradient=False, training=False):
  mean, logvar = model.encode(x, weight=None, reverse_gradient = reverse_gradient, training=training)
  z = model.reparameterize(mean, logvar)
  z_all = [z]
  for t in range(model.num_steps-1):
      z_all.append(tf.zeros((z.shape)))
  z_all = tf.stack(z_all, axis = 1)

  behaviour_preds = model.behaviour_decoder(z_all, training=training)
  total_b_diff = 0
  for t in range(x.shape[1]):
         diff_t = tf.square(tf.cast(gt_behaviour[:,t,:], tf.float32) - model.behaviour_factor_decoder(behaviour_preds[:,t,:], training=training))
         total_b_diff += tf.reduce_sum(diff_t, axis = [1])
         
  behaviour_loss = tf.reduce_mean(total_b_diff)

  loss_b = behaviour_loss#+ 0.01*kl_loss
  return loss_b

def train_kl(model, x, kl_weight, reverse_gradient=False, training=False):
  mean, logvar = model.encode(x, weight=None, reverse_gradient = reverse_gradient, training=training)
  

  kl_loss = - 0.5 * tf.reduce_mean(logvar - tf.square(mean) - tf.exp(logvar) + 1)                  

  loss_kl = kl_weight*kl_loss
  return loss_kl



def train_spikes(model, x, rev_weight, reverse_gradient=True, training=False):
  mean, logvar = model.encode(x, weight =rev_weight, reverse_gradient = reverse_gradient, training=training)
  z = model.reparameterize(mean, logvar)  
  z_all = [z]
  for t in range(model.num_steps-1):
      z_all.append(tf.zeros((z.shape)))
  z_all = tf.stack(z_all, axis = 1)
 
  x_logit = model.decode(z_all, training=training)
  total_recon_loss = 0
  for t in range(x.shape[1]):
       loglikelihood_t = Poisson(model.rates(x_logit[:,t,:], training=training)).logp(x[:,t,:])
       total_recon_loss += tf.reduce_sum(loglikelihood_t, axis = [1])
  logpx_z = -tf.reduce_mean(total_recon_loss)

                 
  loss_s = logpx_z
  
  return loss_s


def train_step(model, x, gt_behaviour, kl_weight, rev_weight, optimizer_b, optimizer_s, optimizer_kl, train_behaviour_decoder):
  """Executes one training step and returns the loss.

  This function computes the loss and gradients, and uses the latter to
  update the model's parameters.
  """
  with tf.GradientTape() as tape_kl:
    loss_kl = train_kl(model, x, kl_weight, reverse_gradient=False, training=True)
      
  gradients_kl = tape_kl.gradient(loss_kl, model.trainable_variables)
  optimizer_kl.apply_gradients([(grad, var) for (grad, var) in zip(gradients_kl, model.trainable_variables) if grad is not None])

  
  with tf.GradientTape() as tape_b:
    loss_b = train_behaviour(model, x, gt_behaviour, reverse_gradient=False, training=True)
      
  gradients_b = tape_b.gradient(loss_b, model.trainable_variables)
  optimizer_b.apply_gradients([(grad, var) for (grad, var) in zip(gradients_b, model.trainable_variables) if grad is not None])
  
  with tf.GradientTape() as tape_s:
    loss_s = train_spikes(model, x, rev_weight, reverse_gradient=True, training=True)
    
  gradients_s = tape_s.gradient(loss_s, model.trainable_variables)
  optimizer_s.apply_gradients([(grad, var) for (grad, var) in zip(gradients_s, model.trainable_variables) if grad is not None])
  

  

def load_data(train_files, test_files):
    for f in train_files:
        print(f['train_data'].shape)
    len_trial = min([f['train_data'].shape[1] for f in train_files + test_files])
    num_neurons = min([f['train_data'].shape[2] for f in train_files + test_files])
    
    train_data = np.concatenate(np.array([f['train_data'][:, :len_trial, :num_neurons] for f in train_files]))
    valid_data = np.concatenate(np.array([f['valid_data'][:, :len_trial, :num_neurons] for f in train_files]))
    test_data = np.concatenate(np.array([f['test_data'][:, :len_trial, :num_neurons] for f in train_files]))
    train_data_all = np.concatenate((train_data, valid_data, test_data))
    
    train_data = np.concatenate(np.array([f['train_data'][:, :len_trial, :num_neurons] for f in test_files]))
    valid_data = np.concatenate(np.array([f['valid_data'][:, :len_trial, :num_neurons] for f in test_files]))
    test_data = np.concatenate(np.array([f['test_data'][:, :len_trial, :num_neurons] for f in test_files]))
    test_data_all = np.concatenate((train_data, valid_data, test_data))
    
    
    train_target = np.concatenate(np.array([f['train_target_direction'][:] for f in train_files]))    
    valid_target = np.concatenate(np.array([f['valid_target_direction'][:] for f in train_files]))
    test_target = np.concatenate(np.array([f['test_target_direction'][:] for f in train_files]))
    train_target_all = np.concatenate((train_target, valid_target, test_target))
    
    train_target = np.concatenate(np.array([f['train_target_direction'][:] for f in test_files]))    
    valid_target = np.concatenate(np.array([f['valid_target_direction'][:] for f in test_files]))
    test_target = np.concatenate(np.array([f['test_target_direction'][:] for f in test_files]))
    test_target_all = np.concatenate((train_target, valid_target, test_target))
    

    mean_xy = np.array([0,0])
    offsets = []
    for m in train_files:
        offsets.append(np.mean(mean_xy - m['train_behaviours'][:, 0], axis = 0))
    train_behaviours = np.concatenate(np.array([f['train_behaviours'][:, :len_trial] + np.array(offsets)[i] for i, f in enumerate(train_files)]))
    valid_behaviours = np.concatenate(np.array([f['valid_behaviours'][:, :len_trial]  + np.array(offsets)[i] for i, f in enumerate(train_files)]))
    test_behaviours = np.concatenate(np.array([f['test_behaviours'][:, :len_trial]  + np.array(offsets)[i] for i, f in enumerate(train_files)]))
    train_behaviours_all = np.concatenate((train_behaviours, valid_behaviours, test_behaviours))
    
    offsets = []
    for m in test_files:
        offsets.append(np.mean(mean_xy - m['train_behaviours'][:, 0], axis = 0))
    train_behaviours = np.concatenate(np.array([f['train_behaviours'][:, :len_trial] + np.array(offsets)[i] for i, f in enumerate(test_files)]))
    valid_behaviours = np.concatenate(np.array([f['valid_behaviours'][:, :len_trial]  + np.array(offsets)[i] for i, f in enumerate(test_files)]))
    test_behaviours = np.concatenate(np.array([f['test_behaviours'][:, :len_trial]  + np.array(offsets)[i] for i, f in enumerate(test_files)]))
    test_behaviours_all = np.concatenate((train_behaviours, valid_behaviours, test_behaviours))
    
    return train_data_all, test_data_all, train_target_all, test_target_all, train_behaviours_all, test_behaviours_all




def plot_directions(model, data, ground_truth_behaviours, ground_truth_directions, save_name, latent_dim, batch_size, num_factors):
    # mean, logvar = model.encode(data[:data.shape[0]//3])
    # z1 = model.reparameterize(mean, logvar)
    # mean, logvar = model.encode(data[(data.shape[0]//3):(data.shape[0]//3)*2])
    # z2 = model.reparameterize(mean, logvar)
    # mean, logvar = model.encode(data[((data.shape[0]//3)*2):])
    # z3 = model.reparameterize(mean, logvar)
    # z = np.concatenate((z1, z2, z3))
    z = np.zeros((data.shape[0], latent_dim))
    #data_split = np.array_split(data, data.shape[0]//batch_size)
    for chunk in range(data.shape[0]//batch_size):           
        mean, logvar = model.encode(data[chunk*batch_size: (chunk + 1)*(batch_size)])
        z[chunk*batch_size: (chunk + 1)*(batch_size)] = model.reparameterize(mean, logvar)
    z_all = [z]
    for t in range(model.num_steps-1):
        z_all.append(tf.zeros((z.shape)))
    z_all = tf.stack(z_all, axis = 1)
    
    
    predicted_behaviours = np.zeros((data.shape[0], data.shape[1], num_factors))
    #data_split = np.array_split(z_all, z_all.shape[0]//batch_size)
    for chunk in range(data.shape[0]//batch_size):                    
        predicted_behaviours[chunk*batch_size: (chunk + 1)*(batch_size)] = model.behaviour_decoder(z_all[chunk*batch_size: (chunk + 1)*(batch_size)])
    #bp1 = model.behaviour_decoder(z_all[:(z_all.shape[0]//2)])
    #bp2 = model.behaviour_decoder(z_all[(z_all.shape[0]//2):])
    #predicted_behaviours = np.concatenate((bp1, bp2))
    #predicted_behaviours = model.behaviour_decoder(z_all)
    final_predicted = np.zeros((data.shape[0], data.shape[1], 2))
    for t in range(model.num_steps):
        final_predicted[:,t,:] = model.behaviour_factor_decoder(predicted_behaviours[:,t,:])
    dir_index = [sorted(set(ground_truth_directions)).index(i) for i in ground_truth_directions]
    colors = plt.cm.nipy_spectral(np.arange(8)/8)
    plt.figure(figsize=(5,5))
    for t in range(final_predicted.shape[0]):
        plt.plot(ground_truth_behaviours[t,:,0],ground_truth_behaviours[t,:,1],color=colors[dir_index[t]],alpha=.5, ls='--')
        plt.plot(final_predicted[t,:,0],final_predicted[t,:,1],color=colors[dir_index[t]],alpha=.5,lw=2)
    plt.title('behaviour')
    plt.legend(('true','model'))
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.xticks(())
    plt.yticks(())
    plt.savefig(save_name + '.pdf')
    plt.clf()

def rad_to_one_hot(data):
    for i, d in enumerate(data):
        data[i] = (math.degrees(d) + 135)/45
    return data

def rad_to_num(rads):
    return (math.degrees(rads) + 135)/45
class Poisson(object):
    """Poisson distributon

    Computes the log probability under the model.

    """

    def __init__(self, log_rates):
        """ Create Poisson distributions with log_rates parameters.

        Args:
          log_rates: a tensor-like list of log rates underlying the Poisson dist.
        """
        self.logr = log_rates

    def logp(self, bin_counts):
        """Compute the log probability for the counts in the bin, under the model.

        Args:
          bin_counts: array-like integer counts

        Returns:
          The log-probability under the Poisson models for each element of
          bin_counts.
        """
        k = tf.cast(bin_counts, tf.float32)
        # log poisson(k, r) = log(r^k * e^(-r) / k!) = k log(r) - r - log k!
        # log poisson(k, r=exp(x)) = k * x - exp(x) - lgamma(k + 1)
        return k * self.logr - tf.exp(self.logr) - tf.math.lgamma(k + 1)

def kl_weighting(start, end, rate, epoch):
    return np.float32(min(start * (1 + rate) ** epoch, end))


def rev_weightinig(start, end, rate, epoch):
    return np.float32(max(start * np.exp(-rate * epoch), end))
